import { Result } from "../util/result";
import { RectangularStructure } from "./rectangular_structure";
import { RectangularStructureParams } from "./rectangular_structure_params";
import { Structure, StructureFlowMode, StructureFlowRegime } from "./structure";

export { RectangularStructureParams };

/**
 * Equation Cunge80
 */
export class StructureCunge80 extends RectangularStructure {
    /**
     * Calcul du débit avec l'équation Cunge80
     * @param sVarCalc Variable à calculer (doit être égale à Q ici)
     */
    public Equation(sVarCalc: string): Result {
        super.CheckEquation(sVarCalc);
        const data = super.defaultResultData();
        let v: number;

        switch (data.Regime) {
            case StructureFlowRegime.FREE:
                if (data.Mode === StructureFlowMode.WEIR) {
                    const R32: number = 3 * Math.sqrt(3) / 2;
                    v = this.prms.Cd.v * this.prms.L.v * Structure.R2G / R32 * Math.pow(this.prms.h1.v, 1.5);
                    this.debug("StructureCunge80.Equation WEIR FREE Q=" + v);
                } else {
                    v = this.prms.Cd.v * this.prms.L.v * Structure.R2G
                        * this.prms.W.v * Math.pow(this.prms.h1.v - this.prms.W.v, 0.5);
                    this.debug("StructureCunge80.Equation ORIFICE FREE Q=" + v);
                }
                break;
            case StructureFlowRegime.SUBMERGED:
                if (data.Mode === StructureFlowMode.WEIR) {
                    v = this.prms.Cd.v * this.prms.L.v * Structure.R2G * this.prms.h2.v
                        * Math.sqrt(this.prms.h1.v - this.prms.h2.v);
                    this.debug("StructureCunge80.Equation WEIR SUBMERGED Q=" + v);
                } else {
                    v = this.prms.Cd.v * this.prms.L.v * Structure.R2G
                        * this.prms.W.v * Math.sqrt(this.prms.h1.v - this.prms.h2.v);
                    this.debug("StructureCunge80.Equation ORIFICE SUBMERGED Q=" + v);
                }
        }

        return new Result(v, data);
    }

    protected getFlowRegime() {
        if (this.prms.h1.v >= 1.5 * this.prms.h2.v) {
            return StructureFlowRegime.FREE;
        } else {
            return StructureFlowRegime.SUBMERGED;
        }
    }

    protected getFlowMode() {
        const regime: StructureFlowRegime = this.getFlowRegime();
        switch (regime) {
            case StructureFlowRegime.FREE:
                if (this.prms.W.v <= 2 / 3 * this.prms.h1.v) {
                    return StructureFlowMode.ORIFICE;
                } else {
                    return StructureFlowMode.WEIR;
                }
            case StructureFlowRegime.SUBMERGED:
                if (this.prms.W.v <= this.prms.h2.v) {
                    return StructureFlowMode.ORIFICE;
                } else {
                    return StructureFlowMode.WEIR;
                }
        }
    }
}
