import { cLog } from "./log";
import { Result } from "./result";
import { MessageSeverity, Message } from "./message";

/**
 * Résultat global d'un calcul
 * Peut comporter un ou plusieurs Result.
 */
// tslint:disable-next-line:max-classes-per-file
export class CalcResult {
    /**
     * Messages (erreurs, infos, ...)
     */
    private _log: cLog;

    private _results: Result[];

    constructor() {
        this._log = new cLog();
        this._results = [];
    }

    public addResult(r: Result) {
        this._results.push(r);
    }

    /**
     * Insert result r at index i
     * @param r Result to insert
     * @param i Index position
     */
    public insertResult(r: Result, i: number) {
        this._results.splice(i, 0, r);
    }

    /**
     * @param name nom du résultat complémentaire
     * @returns le nombre la taille d'un résultat complémentaire
     */
    public extraResultLength(name: string): number {
        const r = this.getExtraResult(name);
        if (r === undefined) {
            return 0;
        }
        return Object.keys(r).length;
    }

    public getExtraResult(name: string): any {
        const res = [];
        for (const r of this._results) {
            const er = r.getExtraResult(name);
            if (er !== undefined) {
                res.push(er);
            }
        }

        switch (res.length) {
            case 0:
                return undefined;

            case 1:
                return res[0];

            default:
                throw new Error(
                    "CalcResult.getExtraResult() : il existe plusieurs 'Result' avec le nom '" + name + "'",
                );
        }
    }

    public get log() {
        return this._log;
    }

    protected get messageCount(): number {
        return this._log.messages.length;
    }

    /**
     * ajoute un message au journal
     */
    public addMessage(m: Message) {
        this._log.add(m);
    }

    /**
     * ajoute tous les messages d'un journal au journal
     */
    public addLog(l: cLog) {
        this._log.addLog(l);
    }

    /**
     * insert un message en début de liste
     */
    public insertMessage(m: Message) {
        this._log.insert(m);
    }

    /**
     * Teste si tous les résultats + complémentaires sont valides et journal sans erreur
     * @returns true si le test est valide
     */
    public get ok(): boolean {
        return this.resultsOk && !this.hasErrorMessages;
    }

    /**
     * Teste si tous les résultats + complémentaires sont valides
     * @returns true si le test est valide
     */
    public get resultsOk(): boolean {
        if (this._results.length === 0 && this.log.messages.length === 0) {
            return false;
        }
        for (const r of this._results) {
            if (!r.resultsOk) {
                return false;
            }
        }
        return true;
    }

    /**
     * @returns true si au moins un message de log comporte un code d'erreur
     */
    public get hasErrorMessages(): boolean {
        for (const m of this.log.messages) {
            if (m.getSeverity() === MessageSeverity.ERROR) {
                return true;
            }
        }

        for (const r of this._results) {
            if (r.hasErrorMessages) {
                return true;
            }
        }

        return false;
    }
}
