import { cLog } from "./log";
import { Message, MessageCode, MessageSeverity } from "./message";
import { CalcResult } from "./calcresult";

/**
 * Résultat de calcul comprenant la valeur du résultat et des calculs annexes (flag, calculs intermédiaires...)
 * Représente une valeur unique
 */
// tslint:disable-next-line:max-classes-per-file
export class Result extends CalcResult {
    /**
     * Valeur calculée
     */
    private _vCalc: number;

    /**
     * résultats complémentaires
     */
    private _extraResults: { [key: string]: any };

    constructor(v: number | Message, d?: {}) {
        super();
        if (typeof v === "number") {
            this._vCalc = v;
            const m = new Message(MessageCode.ERROR_OK);
            m.extraVar = d;
            this.addMessage(m);
        } else if (v instanceof Message) {
            if (d !== undefined) {
                throw new Error("appel à Result.constructor() invalide");
            }
            this.addMessage(v);
        }
        this._extraResults = {};
    }

    get vCalc() { return this._vCalc; }

    get code(): MessageCode {
        if (this.messageCount !== 1) {
            throw new Error("appel à Result.code invalide, il n'y pas exactement un message de log");
        }
        return this.log.messages[0].code;
    }

    get extraVar() {
        if (this.messageCount !== 1) {
            throw new Error("appel à Result.extraVar invalide, il n'y pas exactement un message de log");
        }
        return this.log.messages[0].extraVar;
    }

    private get hasExtraResults(): boolean {
        return Object.keys(this._extraResults).length > 0;
    }

    public get resultsOk(): boolean {
        return this._vCalc !== undefined || this.hasExtraResults;
    }

    public get hasErrorMessages(): boolean {
        for (const m of this.log.messages) {
            if (m.code !== MessageCode.ERROR_OK && m.getSeverity() === MessageSeverity.ERROR) {
                return true;
            }
        }
        return false;
    }

    /**
     * résultats complémentaires
     */

    public get extraResults() {
        return this._extraResults;
    }

    public addExtraResult(name: string, value: any) {
        this._extraResults[name] = value;
    }

    /**
     * @param name nom du résultat complémentaire à retourner
     * @returns un résultat complémentaire par son nom
     */
    public getExtraResult(name: string): any {
        for (const n in this._extraResults) {
            if (n === name) {
                return this._extraResults[n];
            }
        }
        return undefined;
    }

    public toString(): string {
        if (this._vCalc !== undefined) {
            return String(this._vCalc);
        }
        return JSON.stringify(this);
    }
}
